//===========================================================================================//
// This software is based on Robert Walter's code (github.com/robbeofficial)                 //
// Usage:                                                                                    //
//      1) $./NiCapture_13                                                                   //
//      2) Insert the capture name using the format CX_Y_ where: X = user ID, Y = capture ID //
//      3) Start to capture                                                                  //
// Controls during the capture:                                                              //
//      Key 's': start/stop recording                                                        //
//      Key 'd': go back to the previous hand configuration                                  //
//      Key ',': proceed to the next hand configuration                                      //
//===========================================================================================//

#include <iostream>
#include <sstream>
#include <fstream>
using namespace std;

#include <opencv/cv.h>
#include <opencv/highgui.h>
using namespace cv;

#include "functions.h"

#include <XnCppWrapper.h>
using namespace xn;

/*//////////////////////////////////////////////////////////////////////
// constants
//////////////////////////////////////////////////////////////////////*/

const Size frameSize(640, 480);
const unsigned int maxUsers = 20;
const Mat emptyMat();

/*//////////////////////////////////////////////////////////////////////
// globals
//////////////////////////////////////////////////////////////////////*/

//Recorder
xn::Recorder record;
ofstream skelFile;

//to read bones from file..
ifstream skelInfo;

// openNI context and generator nodes
Context context;
ImageGenerator imageGen;
DepthGenerator depthGen;
IRGenerator irGen;
UserGenerator userGen;

// frame buffers
Mat bgrMat(frameSize, CV_8UC3);
Mat depthMat(frameSize, CV_16UC1);
Mat depthMat8(frameSize, CV_8UC1);
Mat depthMatBgr(frameSize, CV_8UC3);

/*//////////////////////////////////////////////////////////////////////
// callbacks
//////////////////////////////////////////////////////////////////////*/

void XN_CALLBACK_TYPE onNewUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
	cout << "onNewUser(" << nId << ")" << endl;
	SkeletonCapability skelCap = userGen.GetSkeletonCap();

	// load default calibration data
	skelCap.LoadCalibrationDataFromFile(nId, "UserCalibration.bin");

	// instantly track skeleton
	skelCap.StartTracking(nId);
}

void XN_CALLBACK_TYPE onLostUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
	printf("Lost user %d\n", nId);
}

/*//////////////////////////////////////////////////////////////////////
// functions
//////////////////////////////////////////////////////////////////////*/

const Scalar skeletonColor(255,255,255);

void drawBone(XnUserID player,XnSkeletonJoint eJoint1, XnSkeletonJoint eJoint2)
{
	XnSkeletonJointPosition joint1, joint2;
	userGen.GetSkeletonCap().GetSkeletonJointPosition(player, eJoint1, joint1);
	userGen.GetSkeletonCap().GetSkeletonJointPosition(player, eJoint2, joint2);

	if (joint1.fConfidence < 0.5 || joint2.fConfidence < 0.5)
	{
		return;
	}

	XnPoint3D pt[2];
	pt[0] = joint1.position;
	pt[1] = joint2.position;

	depthGen.ConvertRealWorldToProjective(2, pt, pt);

	line(depthMatBgr, Point(pt[0].X, pt[0].Y), Point(pt[1].X, pt[1].Y), skeletonColor);
}

void writeBone(XnUserID player,XnSkeletonJoint eJoint)
{
    XnSkeletonJointPosition joint;
	userGen.GetSkeletonCap().GetSkeletonJointPosition(player, eJoint, joint);

	if (joint.fConfidence < 0.5)
	{
		skelFile << "1000 1000 ";
	}
	else
	{
	    XnPoint3D pt[1];
        pt[0] = joint.position;
        depthGen.ConvertRealWorldToProjective(1, pt, pt);
        skelFile << pt[0].X << " " << pt[0].Y << " ";
	}
}

void drawSkeleton(XnUserID player)
{
	drawBone(player, XN_SKEL_HEAD, XN_SKEL_NECK);

	drawBone(player, XN_SKEL_NECK, XN_SKEL_LEFT_SHOULDER);
	drawBone(player, XN_SKEL_LEFT_SHOULDER, XN_SKEL_LEFT_ELBOW);
	drawBone(player, XN_SKEL_LEFT_ELBOW, XN_SKEL_LEFT_HAND);

	drawBone(player, XN_SKEL_NECK, XN_SKEL_RIGHT_SHOULDER);
	drawBone(player, XN_SKEL_RIGHT_SHOULDER, XN_SKEL_RIGHT_ELBOW);
	drawBone(player, XN_SKEL_RIGHT_ELBOW, XN_SKEL_RIGHT_HAND);

	drawBone(player, XN_SKEL_LEFT_SHOULDER, XN_SKEL_TORSO);
	drawBone(player, XN_SKEL_RIGHT_SHOULDER, XN_SKEL_TORSO);
}

void drawBone(XnPoint3D pt1, XnPoint3D pt2)
{
    line(depthMatBgr, Point(pt1.X, pt1.Y), Point(pt2.X, pt2.Y), skeletonColor);
}

void drawSkeletonFromFile(int currentFrame)
{
    XnPoint3D pt[24];

    std::string line;
    getline (skelInfo,line);

    stringstream ss(line);
    string s;

    getline(ss, s, ' ');//frame

    getline(ss, s, ' ');
    pt[XN_SKEL_HEAD].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_HEAD].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_NECK].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_NECK].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_TORSO].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_TORSO].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_SHOULDER].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_SHOULDER].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_ELBOW].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_ELBOW].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_HAND].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_HAND].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_SHOULDER].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_SHOULDER].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_ELBOW].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_ELBOW].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_HAND].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_HAND].Y = atoi(s.c_str());

    drawBone(pt[XN_SKEL_HEAD], pt[XN_SKEL_NECK]);

	drawBone(pt[XN_SKEL_NECK], pt[XN_SKEL_LEFT_SHOULDER]);
	drawBone(pt[XN_SKEL_LEFT_SHOULDER], pt[XN_SKEL_LEFT_ELBOW]);
	drawBone(pt[XN_SKEL_LEFT_ELBOW], pt[XN_SKEL_LEFT_HAND]);

	drawBone(pt[XN_SKEL_NECK], pt[XN_SKEL_RIGHT_SHOULDER]);
	drawBone(pt[XN_SKEL_RIGHT_SHOULDER], pt[XN_SKEL_RIGHT_ELBOW]);
	drawBone(pt[XN_SKEL_RIGHT_ELBOW], pt[XN_SKEL_RIGHT_HAND]);

	drawBone(pt[XN_SKEL_LEFT_SHOULDER], pt[XN_SKEL_TORSO]);
	drawBone(pt[XN_SKEL_RIGHT_SHOULDER], pt[XN_SKEL_TORSO]);
}

void writeSkeleton(XnUserID player, int currentFrame)
{
    skelFile << convertInt(currentFrame) << " ";

    writeBone(player,  XN_SKEL_HEAD);

	writeBone(player, XN_SKEL_NECK);
	writeBone(player, XN_SKEL_TORSO);

	writeBone(player, XN_SKEL_LEFT_SHOULDER);
	writeBone(player, XN_SKEL_LEFT_ELBOW);
    writeBone(player, XN_SKEL_LEFT_HAND);

	writeBone(player, XN_SKEL_RIGHT_SHOULDER);
	writeBone(player, XN_SKEL_RIGHT_ELBOW);
	writeBone(player, XN_SKEL_RIGHT_HAND);

	skelFile << "\n";
}


void initKinect(bool initImage, bool initDepth, bool initIR, bool initUser, char* oniFile) {
	XnStatus nRetVal = XN_STATUS_OK;

	// Initialize context object
	nRetVal = context.Init();
	if(oniFile != "") context.OpenFileRecording(oniFile);
	cout << "init : " << xnGetStatusString(nRetVal) << endl;

	// default output mode
	XnMapOutputMode outputMode;
	outputMode.nXRes = 640;
	outputMode.nYRes = 480;
	outputMode.nFPS = 30;

	// Create an ImageGenerator node
	if (initImage) {
		nRetVal = imageGen.Create(context);
		cout << "imageGen.Create : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = imageGen.SetMapOutputMode(outputMode);
		cout << "imageGen.SetMapOutputMode : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = imageGen.GetMirrorCap().SetMirror(true);
		cout << "imageGen.GetMirrorCap().SetMirror : " << xnGetStatusString(nRetVal) << endl;
	}

	// Create a DepthGenerator node
	if (initDepth) {
		nRetVal = depthGen.Create(context);
		cout << "depthGen.Create : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = depthGen.SetMapOutputMode(outputMode);
		cout << "depthGen.SetMapOutputMode : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = depthGen.GetMirrorCap().SetMirror(true);
		cout << "depthGen.GetMirrorCap().SetMirror : " << xnGetStatusString(nRetVal) << endl;
	}

	// Create an IRGenerator node
	if (initIR) {
		nRetVal = irGen.Create(context);
		cout << "irGen.Create : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = irGen.SetMapOutputMode(outputMode);
		cout << "irGen.SetMapOutputMode : " << xnGetStatusString(nRetVal) << endl;
	}

	// create user generator
	if (initUser) {
		nRetVal = userGen.Create(context);
		cout << "userGen.Create : " << xnGetStatusString(nRetVal) << endl;
		userGen.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL);
	}

	// Make it start generating data
	nRetVal = context.StartGeneratingAll();
	cout << "context.StartGeneratingAll : " << xnGetStatusString(nRetVal) << endl;
}

float getJointImgCoordinates(const SkeletonCapability &skeletonCapability, const XnUserID userId, const XnSkeletonJoint skeletonJoint, float *v) {
	XnVector3D projective;
	XnSkeletonJointPosition skeletonJointPosition;
	skeletonCapability.GetSkeletonJointPosition(userId, skeletonJoint, skeletonJointPosition);

	depthGen.ConvertRealWorldToProjective(1, &skeletonJointPosition.position, &projective);

	v[0] = projective.X;
	v[1] = projective.Y;
	v[2] = projective.Z / 1000.0f;

	return skeletonJointPosition.fConfidence;
}

bool getHandContour(const Mat &depthMat, Mat &bgrMat, const float *v, vector<Point> &handContour) {
	const int maxHandRadius = 200-v[2]*100; // in px

	//cout << maxHandRadius << endl;

	const short handDepthRange = 200; // in mm
	const double epsilon = 17.5; // approximation accuracy (maximum distance between the original hand contour and its approximation)

	unsigned short depth = (unsigned short) (v[2] * 1000.0f); // hand depth
	unsigned short near = depth - 100; // near clipping plane
	unsigned short far = depth + 100; // far clipping plane

	static Mat mask(frameSize, CV_8UC1);
	mask.setTo(0);

	// extract hand region
	circle(mask, Point(v[0], v[1]), maxHandRadius, 255, CV_FILLED);
	mask = mask & depthMat > near & depthMat < far;

	//mask = mask & bgrMat > near & bgrMat < far;

	// DEBUG(show mask)
	//imshow("mask1", mask);

	// assume largest contour in hand region to be the hand contour
	vector<vector<Point> > contours;
	findContours(mask, contours, CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE);
	//findContours(bgrMat, contours, CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE);

	int n = contours.size();
	int maxI = -1;
	int maxSize = -1;
	for (int i=0; i<n; i++) {
		int size  = contours[i].size();
		if (size > maxSize) {
			maxSize = size;
			maxI = i;
		}
	}


    if(maxI < contours.size())
    for(int j=0;j<contours[maxI].size();j++){
        Point p = contours[maxI][j];

        int offsetX = (p.x - 430)*0.08;

        //p.x -= offsetX;

        int offsetY = (p.y - 400)*0.05;

        //p.y -= offsetY;

        line(bgrMat,p,p,cvScalar(255,0,0),3);
    }



	bool handContourFound = (maxI >= 0);

	if (handContourFound) {
		approxPolyDP( Mat(contours[maxI]), handContour, epsilon, true );
		//handContour = contours[maxI];
	}

	// DEBUG(draw hand contour
	mask.setTo(0);
	if (maxI >= 0) {
		vector<vector<Point> > contours2;
		contours2.push_back(contours[maxI]);
		drawContours(mask, contours2, -1, 255);
		//imshow("mask2", mask);
	}

	return maxI >= 0;
}

void detectFingerTips(const vector<Point> &handContour, vector<Point> &fingerTips, Mat *debugFrame = NULL) {
	Mat handContourMat(handContour);
	double area = cv::contourArea(handContourMat);

	const Scalar debugFingerTipColor(255,0,0);

	vector<int> hull;
	cv::convexHull(handContourMat, hull);

	// find upper and lower bounds of the hand and define cutoff threshold (don't consider lower vertices as fingers)
	int upper = 640, lower = 0;
	for (int j=0; j<hull.size(); j++) {
		int idx = hull[j]; // corner index
		if (handContour[idx].y < upper) upper = handContour[idx].y;
		if (handContour[idx].y > lower) lower = handContour[idx].y;
	}
	float cutoff = lower - (lower - upper) * 0.1f;

	// find interior angles of hull corners
	for (int j=0; j<hull.size(); j++) {
		int idx = hull[j]; // corner index
		int pdx = idx == 0 ? handContour.size() - 1 : idx - 1; //  predecessor of idx
		int sdx = idx == handContour.size() - 1 ? 0 : idx + 1; // successor of idx

		Point v1 = handContour[sdx] - handContour[idx];
		Point v2 = handContour[pdx] - handContour[idx];

		float angle = acos( (v1.x*v2.x + v1.y*v2.y) / (norm(v1) * norm(v2)) );

		// low interior angle + within upper 90% of region -> we got a finger
		if (angle < 1 && handContour[idx].y < cutoff) {
			int u = handContour[idx].x;
			int v = handContour[idx].y;

			fingerTips.push_back(Point2i(u,v));

			if (debugFrame) { // draw fingertips
				cv::circle(*debugFrame, handContour[idx], 10, debugFingerTipColor, -1);
			}
		}
	}

	if (debugFrame) {
		// draw cutoff threshold
		cv::line(*debugFrame, Point(0, cutoff), Point(640, cutoff), debugFingerTipColor);

		// draw approxCurve
		for (int j=0; j<handContour.size(); j++) {
			cv::circle(*debugFrame, handContour[j], 10, debugFingerTipColor);
			if (j != 0) {
				cv::line(*debugFrame, handContour[j], handContour[j-1], debugFingerTipColor);
			} else {
				cv::line(*debugFrame, handContour[0], handContour[handContour.size()-1], debugFingerTipColor);
			}
		}

		// draw approxCurve hull
		for (int j=0; j<hull.size(); j++) {
			cv::circle(*debugFrame, handContour[hull[j]], 10, debugFingerTipColor, 3);
			if(j == 0) {
				cv::line(*debugFrame, handContour[hull[j]], handContour[hull[hull.size()-1]], debugFingerTipColor);
			} else {
				cv::line(*debugFrame, handContour[hull[j]], handContour[hull[j-1]], debugFingerTipColor);
			}
		}
	}
}

void drawContour(Mat &img, const vector<Point> &contour, const Scalar &color) {
	vector<vector<Point> > contours;
	contours.push_back(contour);
	drawContours(img, contours, -1, color);
}

double convexity(const vector<Point> &contour) {
	Mat contourMat(contour);

	vector<int> hull;
	convexHull(contourMat, hull);

	int n = hull.size();
	vector<Point> hullContour;

	for (int i=0; i<n; i++) {
		hullContour.push_back(contour[hull[i]]);
	}

	Mat hullContourMat(hullContour);

	return (contourArea(contourMat) / contourArea(hullContourMat));
}

int main(int argc, char **argv) {
	const float minHandExtension = 0.2f; // in meters
	const double grabConvexity = 0.8;

	XnUserID userIds[maxUsers] = {0};
	XnUInt16 nUsers = maxUsers;
	char key = 0;

    bool readBones = false;

	// init context and generators
	if(argc>1)
	{
	    initKinect(true, true, false, true, argv[1]);

        //load bones file
	    string skelFilename = argv[1];
	    skelFilename = skelFilename.substr(0,skelFilename.size()-3);//remove the oni extension
	    skelFilename+="skl";//add skl
	    skelInfo.open(skelFilename.c_str());

        cout << "Read bones from:" << skelFilename << endl;

	    readBones = true;
	}
	else
	  initKinect(true, true, false, true, "");

	// register some handy callbacks
	XnCallbackHandle hUserCallbacks;
	userGen.RegisterUserCallbacks(onNewUser, onLostUser, NULL, hUserCallbacks);

	SkeletonCapability skelCap = userGen.GetSkeletonCap();

	Mat mask(frameSize, CV_8UC1);

	int oldUsers = 0;

	bool recording = false;
    int currentFrame=0;

    setupSamplesWindow();
	loadSigns();

	std::string userName;
	if(argc<=1)
	{
	    cout << "User:";
        cin >> userName;
	}


    //calibrate kinect
    //http://www.informatik.uni-freiburg.de/~engelhar/calibration.html
	CvMat * camera_matrix = (CvMat*)cvLoad("Intrinsics_RGB.xml");
    CvMat * distortion_coeffs = (CvMat*)cvLoad("Distortion_RGB.xml");

    Mat undistortedImg(frameSize, CV_8UC3);

	while ( (key = (char) waitKey(1)) != 27 ) {

		context.WaitAndUpdateAll();

        if(key=='s') {
            if(!recording){

                std::string filename = "database/";
                filename += userName;
                filename += "_";
                filename += convertInt(currentCfg);
                filename += ".oni";

                record.Create(context);
                record.SetDestination(XN_RECORD_MEDIUM_FILE,filename.c_str());
                record.AddNodeToRecording(imageGen,XN_CODEC_JPEG);
                record.AddNodeToRecording(depthGen,XN_CODEC_16Z_EMB_TABLES);
                printf("Start record\n");

                filename = "database/";
                filename += userName;
                filename += "_";
                filename += convertInt(currentCfg);
                filename += ".skl";

                skelFile.open (filename.c_str());

                currentFrame=0;

                recording = true;
            }
            else{
                printf("Stop record\n");
                record.Unref();
                skelFile.close();
                recording = false;
            }
        }

        if(key == 'd')//next hand config
		{
		    currentCfg++;
		    showSample( samples[currentCfg]);
		}

		if(key == ',')//prev hand config
		{
		    if(currentCfg>0) currentCfg--;
		    cout << "Repeat CFG" <<  endl;
		    showSample( samples[currentCfg]);
		}

        if(recording){
            writeSkeleton(userIds[0], currentFrame++);
		}

		// acquire bgr image
		{
			Mat mat(frameSize, CV_8UC3, (unsigned char*) imageGen.GetImageMap());
			cvtColor(mat, bgrMat, CV_RGB2BGR);
		}

		// acquire depth image
		{
			Mat mat(frameSize, CV_16UC1, (unsigned char*) depthGen.GetDepthMap());
			mat.copyTo(depthMat);
			depthMat.convertTo(depthMat8, CV_8UC1, 255.0f / 3000.0f);
			cvtColor(depthMat8, depthMatBgr, CV_GRAY2BGR);
		}

		// iterate over all user
		nUsers = userGen.GetNumberOfUsers();
		userGen.GetUsers(userIds, nUsers);
		float conf; // joint confidence
		float rh[3]; // right hand coordinates (x[px], y[px], z[meters])
		float lh[3]; // left hand coordinates
		float t[3]; // torso coordinates


		for (int i=0; i<nUsers; i++) {
			int id = userIds[i];

			if(skelCap.IsTracking(userIds[i]))
			{
			  drawSkeleton(userIds[i]);
			}

			// torso
			if ( getJointImgCoordinates(skelCap, id, XN_SKEL_TORSO, t) == 1 ) {

				unsigned char shade = 255 - (unsigned char)(t[2] *  128.0f);
				circle(depthMatBgr, Point(t[0], t[1]), 10, Scalar(shade, 0, 0), -1);

				// right hand
				if (
					(getJointImgCoordinates(skelCap, id, XN_SKEL_RIGHT_HAND, rh) == 1) /* confident detection */
					//&& (rh[2] < t[2] - minHandExtension) /* user extends hand towards screen */
					//&& (rh[1] < t[1]) /* user raises his hand */
				) {
					unsigned char shade = 255 - (unsigned char)(rh[2] *  128.0f);
					Scalar color(0, 0, shade);

					vector<Point> handContour;
					getHandContour(depthMat, bgrMat, rh, handContour);
					bool grasp = convexity(handContour) > grabConvexity;
					int thickness = grasp ? CV_FILLED : 3;
					circle(depthMatBgr, Point(rh[0], rh[1]), 10, color, thickness);

					vector<Point> fingerTips;
					detectFingerTips(handContour, fingerTips, &depthMatBgr);
				}

				// left hand
				if (
					(getJointImgCoordinates(skelCap, id, XN_SKEL_LEFT_HAND, lh) == 1)
					&& (lh[2] < t[2] - minHandExtension)
					&& (lh[1] < t[1]) /* user raises his hand */
				) {
					unsigned char shade = 255 - (unsigned char)(lh[2] *  128.0f);
					Scalar color(0, shade, 0);

					vector<Point> handContour;
					getHandContour(depthMat, bgrMat, lh, handContour);
					bool grasp = convexity(handContour) > grabConvexity;
					int thickness = grasp ? CV_FILLED : 3;
					circle(depthMatBgr, Point(lh[0], lh[1]), 10, color, thickness);

					vector<Point> fingerTips;
					detectFingerTips(handContour, fingerTips, &depthMatBgr);
				}
			}
		}

        if(readBones) drawSkeletonFromFile(currentFrame++);

        undistort(bgrMat,undistortedImg,camera_matrix,distortion_coeffs);
        undistortedImg += bgrMat/2;

		imshow("depthMatBgr", depthMatBgr);
		imshow("bgrMat", bgrMat);
		//imshow("undistortedImg",undistortedImg);
	}

    printf("Shutdown...");

    releaseSamplesWindow();
    context.Shutdown();

	return 0;
}
